#!/usr/bin/python
# coding:utf-8
from hbshare.quant.CChen.func import generate_table
from hbshare.quant.CChen.cta_factor.const import composite_factor
from hbshare.quant.CChen.stk import trade_calendar
import pandas as pd
import numpy as np
import pymysql
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from pandas.plotting import register_matplotlib_converters
from datetime import datetime, timedelta

register_matplotlib_converters()
plt.rcParams['font.sans-serif'] = ['SimHei']

sql_l = '''(
    `ID` bigint not null AUTO_INCREMENT,
    `TDATE`  date not null,
    `EXCHANGE`  int,
    `PCODE`  int,
    `UCODE`  int,
    `CLOSE0`  float,
    `CLOSE` float,
    `CLOSE1` float,
    `POS` int,
    `WEIGHT`  double,
    `RETURN` double,
    `FACTORVALUE` double,
    `FACTOR`  varchar(255),
    primary key (`ID`)
    )
    '''

sql_index = '''(
    `ID` bigint not null AUTO_INCREMENT,
    `TDATE`  date not null,
    `CLOSE` float,
    `TURNOVER` float,
    `FACTOR`  varchar(255),
    primary key (`ID`)
    )
    '''

stock_info_file = 'E:\\codes\\hbshare\\hbshare\\quant\\CChen\\data\\wind_edb.csv'


def factor_table_gen(table, sql_info):

    generate_table(
        database='daily_data',
        table=table,
        generate_sql=sql_l,
        sql_ip=sql_info['ip'],
        sql_user=sql_info['user'],
        sql_pass=sql_info['pass'],
        table_comment='CTA factor based on hsjy data'
    )
    print(table + ' generated')


def index_table_gen(table, sql_info):

    generate_table(
        database='daily_data',
        table=table,
        generate_sql=sql_index,
        sql_ip=sql_info['ip'],
        sql_user=sql_info['user'],
        sql_pass=sql_info['pass'],
        table_comment='CTA INDEX'
    )
    print(table + ' generated')


def index_gen_in_loop(data, factor, index0, table, path):
    data['RETURN'] = data['RETURN'] * data['WEIGHT'] * data['POS']
    data_g = data.groupby(by='TDATE').sum()[['RETURN']].reset_index()
    data_g['ret'] = data_g['RETURN'] / 10000 + 1.0
    data['WEIGHT'] = data['WEIGHT'] * data['POS']
    w = data.pivot(index='TDATE', columns='UCODE', values='WEIGHT')
    data_turnover = pd.DataFrame(
        (w.shift(1) - w).abs().sum(axis=1)
    ).rename(columns={0: 'TURNOVER'}).reset_index()
    data_g = data_g.merge(data_turnover, on='TDATE', how='left')
    data_g = data_g.iloc[1:]
    data_g['CLOSE'] = data_g['ret'].cumprod()
    data_g['CLOSE'] = data_g['CLOSE'] * index0
    data_g['FACTOR'] = factor
    data_g[['TDATE', 'CLOSE', 'FACTOR', 'TURNOVER']].to_sql(table, path, if_exists='append', index=False)
    print('\t' + str(len(data_g)) + ', to sql, ' + data_g['TDATE'].tolist()[len(data_g) - 1].strftime('%Y-%m-%d'))


def index_gen(sql_path, sql_info, table_factor='cta_factor', table_index='cta_index'):
    index_table_gen(table=table_index, sql_info=sql_info)
    end_date_all = pd.read_sql_query(
        'select distinct TDATE from ' + table_factor + ' order by TDATE desc limit 1', sql_path
    )
    factors = pd.read_sql_query('select distinct FACTOR from ' + table_factor, sql_path)
    if len(factors) == 0:
        print('No factor return')
        return
    else:
        for i in range(len(factors['FACTOR'])):
            # factor = 'xswr_d1_q75'
            factor = factors['FACTOR'][i]
            last_factor_index = pd.read_sql_query(
                'select * from ' + table_index
                + ' where TDATE<=' + end_date_all['TDATE'][0].strftime('%Y%m%d')
                + ' and FACTOR=\'' + factor
                + '\' order by TDATE desc limit 1',
                sql_path
            )
            if len(last_factor_index) == 0:
                date0 = datetime(1990, 1, 1).date()
                index0 = 1000
            else:
                date0 = last_factor_index['TDATE'][0]
                index0 = last_factor_index['CLOSE'][0]

            print(
                str(i + 1) + '/' + str(len(factors['FACTOR'])) + ', '
                + factor + ', start date: ' + date0.strftime('%Y-%m-%d')
            )

            if date0 == end_date_all['TDATE'][0]:
                print('\tNo new date, ' + date0.strftime('%Y-%m-%d'))
                continue

            factor_data = pd.read_sql_query(
                'select * from ' + table_factor + ' where TDATE>=' + date0.strftime('%Y%m%d')
                + ' and FACTOR=\'' + factor + '\' order by TDATE',
                sql_path
            )
            if len(factor_data) > 0:
                index_gen_in_loop(data=factor_data, factor=factor, index0=index0, table=table_index, path=sql_path)
            else:
                print('\tNo new data')


def index_compose(sql_path, sql_info, table_factor='cta_factor', table_index='cta_index', factors=composite_factor):
    index_table_gen(table=table_index, sql_info=sql_info)
    end_date_all = pd.read_sql_query(
        'select distinct TDATE from ' + table_factor + ' order by TDATE desc limit 1', sql_path
    )
    for i in factors:
        last_factor_index = pd.read_sql_query(
            'select * from ' + table_index
            + ' where  FACTOR=\'' + i + '\' order by TDATE desc limit 1',
            sql_path
        )
        if len(last_factor_index) == 0:
            date0 = datetime(1990, 1, 1).date()
            index0 = 1000
        else:
            date0 = last_factor_index['TDATE'][0]
            index0 = last_factor_index['CLOSE'][0]

        print(
            i + ', start date: ' + date0.strftime('%Y-%m-%d')
        )

        if date0 == end_date_all['TDATE'][0]:
            print('\tNo new date, ' + date0.strftime('%Y-%m-%d'))
            continue

        base_factor_data = pd.read_sql_query(
            'select * from ' + table_factor + ' where TDATE>=' + date0.strftime('%Y%m%d')
            + ' and FACTOR in ' + str(tuple(factors[i])) + ' order by TDATE',
            sql_path
        )
        base_factor_data['WEIGHT'] = base_factor_data['POS'] * base_factor_data['WEIGHT']
        base_factor_data['POS'] = 1

        b_w = base_factor_data.groupby(by=['TDATE', 'UCODE'], as_index=False).mean()[['TDATE', 'UCODE', 'WEIGHT']]
        b_info = base_factor_data.drop_duplicates(subset=['TDATE', 'UCODE'])
        b_data = b_info.drop(columns=['WEIGHT']).merge(b_w, on=['TDATE', 'UCODE'], how='left')
        if len(base_factor_data) > 0:
            index_gen_in_loop(data=b_data, factor=i, index0=index0, table=table_index, path=sql_path)
        else:
            print('\tNo new data')


def get_last_position(table, before_date, factor, sql_path):
    existing_cal = pd.read_sql_query(
        'select distinct `TDATE` from ' + table + ' where `TDATE`<=' + before_date.strftime('%Y%m%d')
        + ' and `FACTOR`=\'' + factor + '\' order by `TDATE` desc limit 1',
        sql_path
    )
    if len(existing_cal) >= 1:
        pos = pd.read_sql_query(
            'select * from ' + table + ' where `TDATE`=' + existing_cal['TDATE'][0].strftime('%Y%m%d')
            + ' and `FACTOR`=\'' + factor + '\'',
            sql_path
        )
        print(factor + ', load last date: ' + existing_cal['TDATE'][0].strftime('%Y-%m-%d'))
        return pos

    else:
        print(factor + ', load last date: None')
        return None


def load_factor(sql_path, factor_name, table='cta_factor'):
    index = factor_name

    data = pd.read_sql_query(
        'select * from ' + table + ' where FACTOR=\'' + index + '\'',
        sql_path
    )
    return data


def factor_summary(data, trade_days=250, show=False):
    data['RETURN'] = data['RETURN'] * data['WEIGHT'] * data['POS']
    data['WEIGHT'] = data['WEIGHT'] * data['POS']
    factor_w = data.pivot(index='TDATE', columns='UCODE', values='WEIGHT').fillna(0)
    factor_data_turnover = pd.DataFrame(
        (factor_w.shift(1) - factor_w).abs().sum(axis=1)
    ).rename(columns={0: 'TURNOVER'}).reset_index()
    factor_data_turnover.loc[0, 'TURNOVER'] = factor_w.abs().sum(axis=1)[0]

    data_g = data.groupby(by='TDATE').sum()[['RETURN']].reset_index()
    data_g = data_g.merge(factor_data_turnover, on='TDATE', how='left')
    # data_g = data.groupby(by='TDATE').mean()[['RETURN']].reset_index()
    data_g['ret'] = data_g['RETURN'] / 10000
    annualized_return_average = (data_g['ret'].mean() + 1) ** trade_days - 1
    data_g['ind'] = (data_g['ret'] + 1).cumprod()
    data_g['max_ind'] = data_g['ind'].cummax()
    cumulative_return = data_g['ind'][len(data_g) - 1] / data_g['ind'][0] - 1.0
    annualized_return = (cumulative_return + 1.0) ** (
            365 / (data_g['TDATE'][len(data_g) - 1] - data_g['TDATE'][0]).days
    ) - 1
    return_sample = (data_g['ind'] / data_g['ind'].shift(1) - 1)[1:]
    sigma = return_sample.std(ddof=1) * np.sqrt(trade_days)

    data_g['dd'] = (data_g['ind'] - data_g['max_ind']) / data_g['max_ind']
    data_g['year'] = data_g['TDATE'].apply(lambda x: x.year)

    range_dates = [data_g['TDATE'][0]]

    year_slice = data_g.drop_duplicates(subset=['year'], keep='last')

    if len(year_slice) > 0:
        range_dates += year_slice['TDATE'].tolist()

    result = {
        '累计收益': cumulative_return,
        '最大回撤': min(data_g['dd'][1:]),
        '年化收益率（头尾）': annualized_return,
        '年化收益率（平均）': annualized_return_average,
        '波动率': sigma,
        '换手率': data_g['TURNOVER'].mean()
    }

    if show:
        print(
            '累计收益： \n\t%.2f%%' % round(cumulative_return * 100, 2)
        )
        print(
            '最大回撤：      \n\t%.2f%%' % round(min(data_g['dd'][1:]) * 100, 2)
        )
        print(
            '年化收益率（头尾）： \n\t%.2f%%' % round(annualized_return * 100, 2)
        )
        print(
            '年化收益率（平均）： \n\t%.2f%%' % round(annualized_return_average * 100, 2)
        )
        print(
            '波动率： \n\t%.2f%%' % round(sigma * 100, 2)
        )
        print(
            '换手率： \n\t%.2f%%' % round(data_g['TURNOVER'].mean(), 2)
        )
        print(
            '收益波动比： \n\t%.2f' % round(annualized_return_average / sigma, 2)
        )
        if min(data_g['dd'][1:]) != 0:
            print(
                '收益回撤比：        \n\t%.2f' % round(-annualized_return_average / min(data_g['dd'][1:]), 2)
            )
        else:
            print(
                '收益回撤比：       \n\tinf'
            )

        print('')
        for i in range(1, len(range_dates)):
            range_return = (
                                   data_g[data_g['TDATE'] == range_dates[i]]['ind'].tolist()[0]
                                   / data_g[data_g['TDATE'] == range_dates[i - 1]]['ind'].tolist()[0]
                           ) - 1
            result[range_dates[i - 1].strftime('%Y/%m/%d') + ' - ' + range_dates[i].strftime('%Y/%m/%d')] = range_return
            print(
                range_dates[i - 1].strftime('%Y/%m/%d') + ' - ' + range_dates[i].strftime('%Y/%m/%d')
                + ':\t收益：%.2f%%' % round(range_return * 100, 2)
            )
    return data_g, result


def factor_plot(data, plot_title, cost=0.0001, size=(10, 6)):
    def to_percent(temp, position):
        return '%.2f' % (100 * temp) + '%'

    fig = plt.figure(figsize=size)
    ax1 = fig.add_subplot(2, 1, 1)

    ax1.plot(data['TDATE'], data['ind'])
    ax2 = ax1.twinx()
    ax2.fill_between(data['TDATE'], data['dd'], color='gray', alpha=0.3)
    plt.gca().yaxis.set_major_formatter(FuncFormatter(to_percent))

    plt.title(plot_title)

    ax3 = fig.add_subplot(2, 1, 2)
    ax3.plot(data['TDATE'], data['ind'])
    ax3.plot(data['TDATE'], ((data['RETURN'] / 10000 - data['TURNOVER'] / 100 * cost) + 1).cumprod())
    ax3.plot(data['TDATE'], ((data['RETURN'] / 10000 - data['TURNOVER'] / 100 * cost * 2) + 1).cumprod())
    ax3.plot(data['TDATE'], ((data['RETURN'] / 10000 - data['TURNOVER'] / 100 * cost * 3) + 1).cumprod())
    ax3.plot(data['TDATE'], ((data['RETURN'] / 10000 - data['TURNOVER'] / 100 * cost * 5) + 1).cumprod())
    plt.legend(['原始', '1倍交易成本', '2倍交易成本', '3倍交易成本', '5倍交易成本'])

    ax4 = ax3.twinx()
    ax4.bar(data['TDATE'], data['TURNOVER'] / 100, color='r', alpha=0.3)
    plt.gca().yaxis.set_major_formatter(FuncFormatter(to_percent))

    plt.show()


def factor_compute(
        window_days_list,
        to_sql_path,
        sql_info,
        start_date=datetime(2010, 1, 1).date(),
        end_date=datetime.now(),
        to_table='cta_factor',
        data=None,
        factor_func=None,
        **kwargs
):
    if 'min_volume' in kwargs:
        min_volume = kwargs['min_volume']
    else:
        min_volume = 10000

    factor_table_gen(table=to_table, sql_info=sql_info)

    def get_existing_date():
        exist_index = get_last_position(
            table=to_table,
            before_date=end_date,
            factor=index_name,
            sql_path=to_sql_path
        )
        if exist_index is not None:
            date0 = exist_index['TDATE'][0]
        else:
            date0 = start_date
        return date0

    for window_days in window_days_list:
        if 'hedge_ratio_list' in kwargs:
            for quantile in kwargs['hedge_ratio_list']:
                if 'name' in kwargs:
                    index_name = 'xs' + kwargs['name'] + '_d' + str(window_days) + '_q' + str(quantile)
                else:
                    index_name = factor_func.__name__ + '_d' + str(window_days) + '_q' + str(quantile)
                start_date0 = get_existing_date()
                if data['TDATE'].sort_values().tolist()[-1] == start_date0:
                    print('\t' + index_name + ', no new data. ' + str(datetime.now()))
                    continue
                data_s = data[data['TDATE'] >= start_date0 - timedelta(days=max(window_days, kwargs['liq_days']) * 2)]
                pos = factor_func(
                    start_date=start_date0,
                    data=data_s,
                    ts_n=window_days,
                    quantile=quantile,
                    min_volume=min_volume,
                    **kwargs
                )
                if pos is not None:
                    pos.to_sql(to_table, to_sql_path, if_exists='append', index=False)
                    print(
                        '\t' + index_name + ' to sql, ' + pos['TDATE'].sort_values().tolist()[-1].strftime('%Y-%m-%d')
                        + '. ' + str(datetime.now())
                    )
                else:
                    print('\t' + index_name + ', no new data. ' + str(datetime.now()))

        else:
            if 'name' in kwargs:
                index_name = 'ts' + kwargs['name'] + '_d' + str(window_days)
            else:
                index_name = factor_func.__name__ + '_d' + str(window_days)
            start_date0 = get_existing_date()
            if data['TDATE'].sort_values().tolist()[-1] == start_date0:
                print('\t' + index_name + ', no new data. ' + str(datetime.now()))
                continue
            data_s = data[data['TDATE'] >= start_date0 - timedelta(days=max(window_days, kwargs['liq_days']) * 2)]
            pos = factor_func(
                start_date=start_date0,
                data=data_s,
                ts_n=window_days,
                min_volume=min_volume,
                **kwargs
            )
            if pos is not None:
                pos.to_sql(to_table, to_sql_path, if_exists='append', index=False)
                print(
                    '\t' + index_name + ' to sql, ' + pos['TDATE'].tolist()[-1].strftime('%Y-%m-%d')
                    + '. ' + str(datetime.now())
                )
            else:
                print('\t' + index_name + ', no new data. ' + str(datetime.now()))


def get_stock_data(
        data_indicator, path, products=None, table='wind_edb', table_p='hsjy_fut_info_p', info_file=None, date_lag=0
):
    stock_data_code = []
    for i in data_indicator:
        if products is not None:
            if i not in products:
                continue
        stock_data_code += data_indicator[i]

    stock_data_raw = pd.read_sql_query(
        'select * from ' + table + ' where code in ' + str(tuple(stock_data_code)).replace(',)', ')'),
        path
    ).pivot(index='t_date', columns='code', values='close').ffill()
    if date_lag > 0:
        stock_data_raw['t_date'] = pd.Series(stock_data_raw.index).shift(-date_lag).values
        stock_data_raw = stock_data_raw.set_index('t_date')
        stock_data_raw = stock_data_raw.iloc[:-date_lag]
    if info_file is None:
        stock_data_info = pd.read_csv(stock_info_file)
    else:
        stock_data_info = pd.read_csv(info_file)

    for i in stock_data_raw.columns:
        stock_data_raw.loc[:, i] = (
                stock_data_raw.loc[:, i] * stock_data_info[stock_data_info['指标ID'] == i]['乘数'].tolist()[0]
        )
    tdate_list = trade_calendar(
        start_date=stock_data_raw.index[0],
        end_date=stock_data_raw.index[len(stock_data_raw) - 1]
    )

    stock_data = pd.DataFrame(
        {
            't_date': tdate_list
        }, index=range(len(tdate_list))
    ).merge(stock_data_raw, on='t_date', how='left').rename(columns={'t_date': 'TDATE'}).set_index('TDATE')

    hsjy_p_info = pd.read_sql_query(
        'select * from ' + table_p, path
    )

    data_stock = pd.DataFrame()
    for i in data_indicator:
        if products is not None:
            if i not in products:
                continue
        data_p = pd.DataFrame(stock_data[data_indicator[i]].dropna(axis=0, how='all').sum(axis=1)).rename(
            columns={0: 'WRQCURRENT'}
        )

        p_info = hsjy_p_info[hsjy_p_info['CODE'] == i]

        data_p['EXCHANGE'] = p_info['EXCHANGE'].tolist()[0]
        data_p['PCODE'] = p_info['PCODE'].tolist()[0]
        data_p['PNAME'] = p_info['PNAME'].tolist()[0]
        data_stock = data_stock.append(data_p.reset_index())
    return data_stock.reset_index(drop=True)


def get_mr_data(start_date, end_date, path, table='hsjy_fut_memberrank', ts_n=365):
    data_mr = pd.read_sql_query(
        'select TDATE, CCODE, INDICATORCODE, MCODE , INDICATORVOL, INDICATORCHG from ' + table
        + ' where TDATE<=' + end_date.strftime('%Y%m%d')
        + ' and TDATE>=' + (start_date - timedelta(days=ts_n)).strftime('%Y%m%d')
        + ' and INDICATORCODE in (3, 4) order by TDATE, CCODE, INDICATORCODE',
        path
    ).drop_duplicates(subset=['TDATE', 'CCODE', 'INDICATORCODE', 'MCODE'])
    data_mr = data_mr.groupby(by=['TDATE', 'CCODE', 'INDICATORCODE'], as_index=False).sum().rename(
        columns={
            'INDICATORCODE': 'SIDE',
            'INDICATORVOL': 'MR',
            'INDICATORCHG': 'MRCHG'
        }
    ).reset_index(drop=True)
    date_2020 = data_mr.index[data_mr['TDATE'] < datetime(2020, 1, 1).date()]
    data_mr.loc[date_2020, ['MR', 'MRCHG']] = data_mr.loc[date_2020, ['MR', 'MRCHG']] / 2
    return data_mr

# if __name__ == '__main__':
#     index_gen()



"""
小市值组合模块
"""
import os
import pandas as pd
import hbshare as hbs
import numpy as np
import datetime
from hbshare.fe.common.util.data_loader import get_trading_day_list

path = r'D:\kevin\risk_model_jy\RiskModel\data'


class Portfolio:
    def __init__(self, trade_date, window=60):
        self.trade_date = trade_date
        self.window = window

    def run(self):
        # universe
        univ = pd.read_csv(os.path.join(path, r'common_data/univ/zzqz/{0}.csv'.format(self.trade_date)),
                           index_col=0)
        univ['ticker'] = univ.index
        univ['ticker'] = univ['ticker'].apply(lambda x: str(x).zfill(6))
        universe = univ['ticker'].tolist()
        # 市净率 & 股息率
        sql_script = "SELECT a.SecuCode, b.TradingDay, b.PB, b.DividendRatio FROM hsjy_gg.LC_DIndicesForValuation b " \
                     "join hsjy_gg.SecuMain a on a.InnerCode = b.InnerCode where a.SecuMarket in (83, 90) and " \
                     "a.SecuCategory = 1 and to_char(b.TradingDay,'yyyymmdd') = '{}'".format(self.trade_date)
        fin_data_main = pd.DataFrame(hbs.db_data_query('readonly', sql_script, page_size=5000)['data'])

        sql_script = "SELECT a.SecuCode, b.TradingDay, b.PB, b.DividendRatioTTM FROM hsjy_gg.LC_STIBDIndiForValue b " \
                     "join hsjy_gg.SecuMain a on a.InnerCode = b.InnerCode where a.SecuMarket in (83, 90) and " \
                     "a.SecuCategory = 1 and to_char(b.TradingDay,'yyyymmdd') = '{}'".format(self.trade_date)
        fin_data_stib = pd.DataFrame(hbs.db_data_query('readonly', sql_script, page_size=5000)['data'])

        fin_data_stib.rename(columns={"DIVIDENDRATIOTTM": "DIVIDENDRATIO"}, inplace=True)
        fin_data = pd.concat([fin_data_main, fin_data_stib])
        fin_data.rename(columns={"SECUCODE": "ticker", "PB": "pb",
                                 "DIVIDENDRATIO": "dividend", "TRADINGDAY": "trade_date"}, inplace=True)
        fin_data = fin_data.set_index('ticker')[['pb', 'dividend']]
        # 市值
        date_market_value = pd.read_json(os.path.join(path, r'common_data/market_value/%s.json' % self.trade_date),
                                         dtype={'ticker': np.str})
        # 波动率
        trade_dt = datetime.datetime.strptime(self.trade_date, '%Y%m%d')
        pre_date = (trade_dt - datetime.timedelta(days=200)).strftime('%Y%m%d')
        trading_day_list = get_trading_day_list(pre_date, self.trade_date)
        pre_date = trading_day_list[-self.window:][0]
        listdir = os.listdir(os.path.join(path, r'common_data/chg_pct'))
        listdir = [x for x in listdir if pre_date <= x.split('.')[0] <= self.trade_date]
        chg_pct = []
        for filename in listdir:
            date_chg_pct = pd.read_csv(os.path.join(path, r'common_data/chg_pct/{}'.format(filename)),
                                       dtype={'ticker': str, "tradeDate": str})
            chg_pct.append(date_chg_pct)
        chg_pct = pd.concat(chg_pct, axis=0)
        chg_pct.rename(columns={'dailyReturnReinv': 'chg_pct', 'tradeDate': 'trade_date'}, inplace=True)

        chg_pct.loc[chg_pct['chg_pct'] < -0.2, 'chg_pct'] = -0.2
        chg_pct.loc[chg_pct['chg_pct'] > 0.2, 'chg_pct'] = 0.2

        stock_return = pd.pivot_table(chg_pct, index='trade_date', columns='ticker', values='chg_pct').sort_index()
        count_df = stock_return.isnull().sum()
        included_list = count_df[count_df <= 20].index.tolist()
        stock_return = stock_return[included_list]
        stock_vol = stock_return.std().to_frame('vol')
        # 申万一级行业
        data = pd.read_json(os.path.join(path, r'common_data/sw_new.json'), dtype={'ticker': str})
        data['outDate'] = data['outDate'].fillna(20991231)
        data['intoDate'] = data['intoDate'].map(int).map(str)
        data['outDate'] = data['outDate'].map(int).map(str)
        industry = data[(data['intoDate'] <= self.trade_date) & (data['outDate'] >= self.trade_date)].drop_duplicates(
            subset='ticker')

        summary_df = fin_data[['pb', 'dividend']].merge(
            date_market_value.set_index('ticker')[['marketValue']], left_index=True, right_index=True).merge(
            stock_vol, left_index=True, right_index=True).merge(
            industry.set_index('ticker')[['industryName1']], left_index=True, right_index=True).reindex(
            universe).reset_index()

        "======================================================筛选====================================================="
        pb_quantile = summary_df.groupby(
            'industryName1').apply(lambda x: x['pb'].quantile(0.3)).to_frame('pb_quantile').reset_index()
        summary_df = summary_df.merge(pb_quantile, on='industryName1')

        # include_1 = summary_df[summary_df['pb'] <= summary_df['pb_quantile']]['ticker'].tolist()
        # include_2 = summary_df[~summary_df['dividend'].isnull()]['ticker'].tolist()
        # include_3 = summary_df[summary_df['vol'] <= summary_df['vol'].quantile(0.5)]['ticker'].tolist()
        # include_4 = summary_df[summary_df['marketValue'] <= summary_df['marketValue'].quantile(0.5)]['ticker'].tolist()

        selected_df = summary_df[summary_df['pb'] <= summary_df['pb_quantile']]
        selected_df = selected_df.dropna(subset=['dividend'])
        selected_df = selected_df[selected_df['vol'] <= selected_df['vol'].quantile(0.5)]
        selected_df = selected_df[selected_df['marketValue'] <= selected_df['marketValue'].quantile(0.3)]

        for col in ['pb', 'dividend', 'marketValue', 'vol']:
            selected_df[col] = (selected_df[col] - selected_df[col].mean()) / selected_df[col].std()

        selected_df['dividend'] = -1 * selected_df['dividend']

        selected_df['score'] = selected_df[['pb', 'dividend', 'marketValue', 'vol']].mean(axis=1)

        selected_df.sort_values(by='score', inplace=True)

        return selected_df


if __name__ == '__main__':
    Portfolio('20220826').run()


"""
套利 + 市场中性回测模块
"""
import numpy as np
from MyUtil.data_loader import get_trading_day_list, get_fund_nav_from_sql
from MyUtil.util_func import cal_annual_return, cal_annual_volatility, cal_max_drawdown, cal_sharpe_ratio
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_daily_nav_from_work
import pandas as pd
import datetime
import riskfolio as rp


class test:
    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date
        self._load_data()

    def _load_data(self):
        arb_dict = {
            "展弘稳进1号1期": "SE8723",
            # "盛冠达股指套利3号": "SGS597",
            "蒙玺分形3号": "SGP290",
            # "稳博中睿6号": "SJB143",
            "稳博中性稳稳系列1号": "SQP800",
            "蒙玺竞起6号": "SSM855",
            "展弘量化精选1号1期": "SSE967",
            "百奕传家一号": "SJS027",
            "平行公理同余4号": "STT643",
            "悬铃C号6期": "SNY736",
        }

        date_list = get_trading_day_list(self.start_date, self.end_date, frequency='week')
        df = get_fund_nav_from_sql(self.start_date, self.end_date, arb_dict).reindex(date_list)
        # 稳博
        df.loc[:"20210618", "稳博中性稳稳系列1号"] = np.NAN
        # 展弘精选
        df.loc[:"20220304", "展弘量化精选1号1期"] = np.NAN
        df.loc["20220902", "展弘量化精选1号1期"] = 1.0522
        df.loc["20220909", "展弘量化精选1号1期"] = 1.0513
        return_df = df.pct_change().dropna(how='all')
        # 蒙玺
        mx_df = pd.read_excel('D:\\蒙玺竞起6号净值.xlsx', sheet_name=0)
        mx_df['trade_date'] = mx_df['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        mx_df.rename(columns={"nav": "蒙玺竞起6号"}, inplace=True)
        tmp = mx_df.set_index('trade_date').reindex(date_list).dropna().pct_change().dropna()
        return_df.loc[tmp.index, "蒙玺竞起6号"] = tmp.loc[tmp.index, '蒙玺竞起6号']
        return_df.loc[:"20210104", "蒙玺分形3号"] = np.NAN
        # 百奕
        return_df.loc[:"20200508", '百奕传家一号'] = np.NAN
        # 悬铃
        return_df.loc[:"20210528", '悬铃C号6期'] = np.NAN

        tmp = get_daily_nav_from_work(self.start_date, self.end_date, {"平行公理同余4号": "STT643"})
        tmp = tmp.reindex(date_list).loc["20220701":].pct_change().dropna()
        return_df.loc[tmp.index, "平行公理同余4号"] = tmp.loc[tmp.index, '平行公理同余4号']

        arb_return = return_df.mean(axis=1)

        # neu_dict = {
        #     "量锐": "SX4966",
        #     "明汯": "SCQ804",
        #     "启林": "SEA766",
        #     "卓识": "SCL316",
        #     "衍复": "SJH864",
        #     "伯兄": "SL3246",
        #     "概率": "SNM976",
        #     "星阔": "SNU704",
        #     "诚奇": "SNR622",
        #     "乾象": "P48467"
        # }
        # neu_df = get_fund_nav_from_sql(self.start_date, self.end_date, neu_dict).reindex(
        #     date_list)
        # neu_return = neu_df.pct_change().dropna(how='all').mean(axis=1)

        data_with_header = pd.read_excel('D:\\量化产品跟踪\\市场中性\\中性-20220909.xlsx', sheet_name='原始净值')
        data = pd.read_excel('D:\\量化产品跟踪\\市场中性\\中性-20220909.xlsx', sheet_name='原始净值', header=1)
        data['t_date'] = data['t_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        data.index = data['t_date']
        cols = data_with_header.columns.tolist()

        data_param = dict()
        type_list = [x for x in cols if not x.startswith('Unnamed')]
        for i in range(len(type_list) - 1):
            s_index, e_index = cols.index(type_list[i]), cols.index(type_list[i + 1])
            data_slice = data[data.columns[s_index: e_index]]
            data_slice = data_slice[data_slice.index >= self.start_date].reindex(date_list)
            data_param[type_list[i]] = data_slice

        nav_df = pd.concat([data_param['高频'], data_param['中低频']], axis=1)
        cols = [x for x in nav_df.columns if x not in ['英旷尊享1号', '天算顺势1号', '茂源资本-巴舍里耶2期', '星阔云起1号市场中性',
                                                       '佳期星际一期', '概率一号', '世纪前沿量化对冲9号', '龙旗紫微量化对冲',
                                                       '量锐18号', '平方和智增1号']]
        nav_df = nav_df[cols]
        neu_return = nav_df.pct_change(limit=1).dropna(how='all').mean(axis=1)

        return_df = arb_return.to_frame('arbitrage').merge(
            neu_return.to_frame('neutral'), left_index=True, right_index=True)
        # nav_df = (1 + return_df).cumprod()

        self.return_df = return_df

    def run(self):
        return_df = self.return_df
        nav_df = (1 + return_df).cumprod()
        # 固定权重再平衡
        nav_df['trade_date'] = nav_df.index
        nav_df['trade_dt'] = nav_df['trade_date'].apply(lambda x: datetime.datetime.strptime(x, "%Y%m%d"))
        nav_df['month'] = nav_df['trade_dt'].apply(lambda x: x.month)
        nav_df['year'] = nav_df['trade_dt'].apply(lambda x: x.year)
        month_end = nav_df[nav_df['month'].shift(-1) != nav_df['month']]['trade_date'].tolist()
        month_end.append("20181228")
        reb_list = sorted(month_end)[::3]

        weight_df = pd.DataFrame(index=reb_list, columns=["arbitrage", "neutral"])
        weight_df.iloc[0] = 0.5
        weight_df.iloc[1] = 0.5

        for i in range(0, len(reb_list) - 1):
            pre_date, t_date = reb_list[i], reb_list[i + 1]
            period_data = return_df[(return_df.index > pre_date) & (return_df.index <= t_date)]
            port = rp.Portfolio(returns=period_data)
            model = 'Classic'  # 可以为Classic (historical), BL (Black Litterman)、FM (Factor Model)或BLFM
            rm = 'MV'  # 风险度量指标，MV表示方差，本工具包共支持13种风险度量指标
            # obj = 'Sharpe'  # 目标函数, 可选有MinRisk, MaxRet, Utility或Sharpe
            hist = True  # 是否基于历史数据计算风险收益
            rf = 0  # 无风险利率
            # l = 0

            method_mu = 'hist'  # 还支持其他方法，详见文档
            method_cov = 'hist'  # 还支持其他方法，详见文档
            port.assets_stats(method_mu=method_mu, method_cov=method_cov)

            # w = port.optimization(model=model, rm=rm, obj=obj, rf=rf, l=l, hist=hist)

            w = port.rp_optimization(model=model, rm=rm, rf=rf, hist=hist)

            weight_df.loc[t_date] = w['weights']

        # weight_df = pd.DataFrame(index=reb_list, columns=["arbitrage", "neutral"])
        # weight_df['arbitrage'] = 0.7
        # weight_df['neutral'] = 0.3

        ret_list = []
        for i in range(len(reb_list) - 1):
            t_date, next_date = reb_list[i], reb_list[i + 1]
            t_weight = weight_df.loc[t_date]
            period_data = return_df[(return_df.index > t_date) & (return_df.index <= next_date)]
            port = t_weight.dot((1 + period_data).cumprod().T)
            port_return = port.pct_change()
            port_return.iloc[0] = port.iloc[0] - 1

            ret_list.append(port_return)

        ret_series = pd.concat(ret_list).sort_index()

        return_df = return_df.merge(ret_series.to_frame('port'), left_index=True, right_index=True)
        nav_df = (1 + return_df).cumprod()

        portfolio_index_df = pd.DataFrame(
            index=nav_df.columns, columns=['超额年化收益', '超额年化波动', '最大回撤', 'Sharpe', '胜率', '平均损益比',
                                           '2019', '2020', '2021', '2022', '2022(年化)'])
        portfolio_index_df.loc[:, '超额年化收益'] = nav_df.pct_change().dropna(how='all').apply(cal_annual_return, axis=0)
        portfolio_index_df.loc[:, '超额年化波动'] = \
            nav_df.pct_change().dropna(how='all').apply(cal_annual_volatility, axis=0)
        portfolio_index_df.loc[:, '最大回撤'] = nav_df.apply(cal_max_drawdown, axis=0)
        portfolio_index_df.loc[:, 'Sharpe'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: cal_sharpe_ratio(x, 0.015), axis=0)
        portfolio_index_df.loc[:, '胜率'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: x.gt(0).sum() / len(x), axis=0)
        portfolio_index_df.loc[:, '平均损益比'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: x[x > 0].mean() / x[x < 0].abs().mean(), axis=0)
        portfolio_index_df['2019'] = (nav_df.loc["20191227"] - 1).tolist()
        portfolio_index_df['2020'] = (nav_df.loc["20201231"] / nav_df.loc["20191227"] - 1).tolist()
        portfolio_index_df['2021'] = (nav_df.loc["20211231"] / nav_df.loc["20201231"] - 1).tolist()
        portfolio_index_df['2022'] = ((nav_df.loc["20220909"] / nav_df.loc["20211231"]) - 1).tolist()
        portfolio_index_df['2022(年化)'] = ((nav_df.loc["20220909"] / nav_df.loc["20211231"]) ** (52 / 35) - 1).tolist()
        portfolio_index_df.index.name = '产品名称'
        portfolio_index_df.reset_index(inplace=True)
        # 格式处理
        portfolio_index_df['超额年化收益'] = portfolio_index_df['超额年化收益'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['超额年化波动'] = portfolio_index_df['超额年化波动'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['最大回撤'] = portfolio_index_df['最大回撤'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['Sharpe'] = portfolio_index_df['Sharpe'].round(2)
        portfolio_index_df['胜率'] = portfolio_index_df['胜率'].apply(lambda x: format(x, '.1%'))
        portfolio_index_df['平均损益比'] = portfolio_index_df['平均损益比'].round(2)
        portfolio_index_df['2019'] = portfolio_index_df['2019'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['2020'] = portfolio_index_df['2020'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['2021'] = portfolio_index_df['2021'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['2022'] = portfolio_index_df['2022'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['2022(年化)'] = portfolio_index_df['2022(年化)'].apply(lambda x: format(x, '.2%'))

        portfolio_index_df.to_clipboard()


if __name__ == '__main__':
    test('20190101', '20220909').run()